# Tuple with all cim-types where cimsparql assumes mRID exists


CIM_TYPES_WITH_MRID = (
    "ACDCConverterDCTerminal",
    "ACLineSegment",
    "ActivePowerLimit",
    "Analog",
    "ApparentPowerLimit",
    "BaseVoltage",
    "Bay",
    "Breaker",
    "BusNameMarker",
    "BusbarSection",
    "CogenerationPlant",
    "ConformLoad",
    "ConformLoadGroup",
    "ConnectivityNode",
    "ControlArea",
    "CoordinateSystem",
    "CsConverter",
    "CurrentLimit",
    "CurrentRelay",
    "CurrentTransformer",
    "DCConvertUnit",
    "DCNode",
    "DCTerminal",
    "DayType",
    "Disconnector",
    "Discrete",
    "EnergyConsumer",
    "EnergySource",
    "FossilFuel",
    "GeographicalRegion",
    "GroundDisconnector",
    "HydroGeneratingUnit",
    "HydroPowerPlant",
    "HydroPump",
    "Line",
    "LinearShuntCompensator",
    "LoadArea",
    "LoadResponseCharacteristic",
    "Location",
    "MarketParticipant",
    "NonConformLoad",
    "NonConformLoadGroup",
    "OperationalLimitSet",
    "OperationalLimitType",
    "PetersonCoil",
    "PhaseTapChangerLinear",
    "PotentialTransformer",
    "PowerTransformer",
    "PowerTransformerEnd",
    "RatioTapChanger",
    "ReactiveCapabilityCurve",
    "RegulatingControl",
    "Season",
    "SeriesCompensator",
    "ShuntCompensator",
    "StaticVarCompensator",
    "SubGeographicalRegion",
    "SubLoadArea",
    "Substation",
    "SynchrocheckRelay",
    "SynchronousMachine",
    "TapChangerControl",
    "Terminal",
    "ThermalGeneratingUnit",
    "TieFlow",
    "TopologicalNode",
    "VoltageLevel",
    "VoltageLimit",
    "VsConverter",
    "WindGeneratingUnit",
)
